@@ -57,7 +57,7 @@ module Data.Semiring
   ) where
 
 import           Control.Applicative (Applicative(..), Const(..), liftA2)
-import           Data.Bits (Bits)
+import           Data.Bits (Bits (xor))
 import           Data.Bool (Bool(..), (||), (&&), otherwise)
 import           Data.Coerce (Coercible, coerce)
 import           Data.Complex (Complex(..))
@@ -87,12 +87,10 @@ import           Data.Maybe (Maybe(..))
 import           Data.Monoid (Ap(..))
 #endif
 #if defined(VERSION_containers)
-#if MIN_VERSION_base(4,7,0)
 import           Data.IntMap (IntMap)
 import qualified Data.IntMap as IntMap
 import           Data.IntSet (IntSet)
 import qualified Data.IntSet as IntSet
-#endif
 import           Data.Map (Map)
 import qualified Data.Map as Map
 #endif
@@ -101,7 +99,7 @@ import           Data.Ord (Ord((<)), (>=))
 import           Data.Ord (Down(..))
 import           Data.Proxy (Proxy(..))
 import           Data.Ratio (Ratio, Rational, (%))
-import           Data.Semigroup.Compat (Semigroup(..))
+import           Data.Semigroup (Semigroup ((<>), stimes))
 #if defined(VERSION_containers)
 import           Data.Set (Set)
 import qualified Data.Set as Set
@@ -129,25 +127,55 @@ import           GHC.Real (Integral, Fractional, Real, RealFrac)
 import qualified GHC.Real as Real
 import           GHC.Show (Show)
 import           Numeric.Natural (Natural)
-import           Prelude (Integer)
+import           Prelude (Integer, Ordering(..), compare, even, quot)
 
 #if !MIN_VERSION_base(4,12,0)
 import           Language.Haskell.TH.Syntax (Q, Dec, Type)
 import qualified Prelude as P
 #endif
 
-#ifdef mingw32_HOST_OS
-#define HOST_OS_WINDOWS 1
-#else
-#define HOST_OS_WINDOWS 0
+#include "HsBaseConfig.h"
+import           System.Posix.Types (
+#ifdef HTYPE_CC_T
+                 CCc(..),
 #endif
-
-#if !HOST_OS_WINDOWS
-import           System.Posix.Types
-  (CCc, CDev, CGid, CIno, CMode, CNlink,
-   COff, CPid, CRLim, CSpeed, CSsize,
-   CTcflag, CUid, Fd)
+#ifdef HTYPE_DEV_T
+                 CDev(..),
 #endif
+#ifdef HTYPE_GID_T
+                 CGid(..),
+#endif
+#ifdef HTYPE_INO_T
+                 CIno(..),
+#endif
+#ifdef HTYPE_MODE_T
+                 CMode(..),
+#endif
+#ifdef HTYPE_NLINK_T
+                 CNlink(..),
+#endif
+#ifdef HTYPE_OFF_T
+                 COff(..),
+#endif
+#ifdef HTYPE_PID_T
+                 CPid(..),
+#endif
+#ifdef HTYPE_RLIM_T
+                 CRLim(..),
+#endif
+#ifdef HTYPE_SPEED_T
+                 CSpeed(..),
+#endif
+#ifdef HTYPE_SSIZE_T
+                 CSsize(..),
+#endif
+#ifdef HTYPE_TCFLAG_T
+                 CTcflag(..),
+#endif
+#ifdef HTYPE_UID_T
+                 CUid(..),
+#endif
+                 Fd(..))
 
 infixl 7 *, `times`
 infixl 6 +, `plus`, -, `minus`
@@ -337,6 +365,20 @@ newtype Mul a = Mul { getMul :: a }
 instance Semiring a => Semigroup (Mul a) where
   Mul a <> Mul b = Mul (a * b)
   {-# INLINE (<>) #-}
+  stimes n x0 = case compare n 0 of
+    LT -> error "stimes: negative multiplier"
+    EQ -> mempty
+    GT -> f x0 n
+      where
+        f x y
+          | even y = f (x `mappend` x) (y `quot` 2)
+          | y == 1 = x
+          | otherwise = g (x `mappend` x) (y `quot` 2) x
+        g x y z
+          | even y = g (x `mappend` x) (y `quot` 2) z
+          | y == 1 = x `mappend` z
+          | otherwise = g (x `mappend` x) (y `quot` 2) (x `mappend` z)
+  {-# INLINE stimes #-}
 
 instance Semiring a => Monoid (Mul a) where
   mempty = Mul one
@@ -392,10 +434,7 @@ newtype Mod2 = Mod2 { getMod2 :: Bool }
     )
 
 instance Semiring Mod2 where
-  -- we inline the definition of 'xor'
-  -- on Bools, since the instance did not exist until
-  -- base-4.7.0.
-  plus (Mod2 x) (Mod2 y) = Mod2 (x /= y)
+  plus (Mod2 x) (Mod2 y) = Mod2 (x `xor` y)
   times (Mod2 x) (Mod2 y) = Mod2 (x && y)
   zero = Mod2 False
   one = Mod2 True
@@ -449,9 +488,7 @@ instance Ring Mod2 where
 --     @'zero' '*' x = x '*' 'zero' = 'zero'@
 
 class Semiring a where
-#if __GLASGOW_HASKELL__ >= 708
   {-# MINIMAL plus, times, (zero, one | fromNatural) #-}
-#endif
   plus  :: a -> a -> a -- ^ Commutative Operation
   zero  :: a           -- ^ Commutative Unit
   zero = fromNatural 0
@@ -467,9 +504,7 @@ class Semiring a where
 --     @'negate' a '+' a = 'zero'@
 
 class Semiring a => Ring a where
-#if __GLASGOW_HASKELL__ >= 708
   {-# MINIMAL negate #-}
-#endif
   negate :: a -> a
 
 -- | Subtract two 'Ring' values. For any type @R@ with
@@ -504,6 +539,72 @@ fromIntegral x
   Instances (base)
 --------------------------------------------------------------------}
 
+instance (Semiring a, Semiring b) => Semiring (a,b) where
+  zero = (zero, zero)
+  one = (one, one)
+  plus (x1, x2) (y1, y2) =
+    (x1 `plus` y1, x2 `plus` y2)
+  times (x1, x2) (y1, y2) =
+    (x1 `times` y1, x2 `times` y2)
+
+instance (Semiring a, Semiring b, Semiring c) => Semiring (a,b,c) where
+  zero = (zero, zero, zero)
+  one = (one, one, one)
+  plus (x1, x2, x3) (y1, y2, y3) =
+    (x1 `plus` y1, x2 `plus` y2, x3 `plus` y3)
+  times (x1, x2, x3) (y1, y2, y3) =
+    (x1 `times` y1, x2 `times` y2, x3 `times` y3)
+
+instance (Semiring a, Semiring b, Semiring c, Semiring d) => Semiring (a,b,c,d) where
+  zero = (zero, zero, zero, zero)
+  one = (one, one, one, one)
+  plus (x1, x2, x3, x4) (y1, y2, y3, y4) =
+    (x1 `plus` y1, x2 `plus` y2, x3 `plus` y3, x4 `plus` y4)
+  times (x1, x2, x3, x4) (y1, y2, y3, y4) =
+    (x1 `times` y1, x2 `times` y2, x3 `times` y3, x4 `times` y4)
+
+instance (Semiring a, Semiring b, Semiring c, Semiring d, Semiring e) => Semiring (a,b,c,d,e) where
+  zero = (zero, zero, zero, zero, zero)
+  one = (one, one, one, one, one)
+  plus (x1, x2, x3, x4, x5) (y1, y2, y3, y4, y5) =
+    (x1 `plus` y1, x2 `plus` y2, x3 `plus` y3, x4 `plus` y4, x5 `plus` y5)
+  times (x1, x2, x3, x4, x5) (y1, y2, y3, y4, y5) =
+    (x1 `times` y1, x2 `times` y2, x3 `times` y3, x4 `times` y4, x5 `times` y5)
+
+instance (Semiring a, Semiring b, Semiring c, Semiring d, Semiring e, Semiring f) => Semiring (a,b,c,d,e,f) where
+  zero = (zero, zero, zero, zero, zero, zero)
+  one = (one, one, one, one, one, one)
+  plus (x1, x2, x3, x4, x5, x6) (y1, y2, y3, y4, y5, y6) =
+    (x1 `plus` y1, x2 `plus` y2, x3 `plus` y3, x4 `plus` y4, x5 `plus` y5, x6 `plus` y6)
+  times (x1, x2, x3, x4, x5, x6) (y1, y2, y3, y4, y5, y6) =
+    (x1 `times` y1, x2 `times` y2, x3 `times` y3, x4 `times` y4, x5 `times` y5, x6 `times` y6)
+
+instance (Semiring a, Semiring b, Semiring c, Semiring d, Semiring e, Semiring f, Semiring g) => Semiring (a,b,c,d,e,f,g) where
+  zero = (zero, zero, zero, zero, zero, zero, zero)
+  one = (one, one, one, one, one, one, one)
+  plus (x1, x2, x3, x4, x5, x6, x7) (y1, y2, y3, y4, y5, y6, y7) =
+    (x1 `plus` y1, x2 `plus` y2, x3 `plus` y3, x4 `plus` y4, x5 `plus` y5, x6 `plus` y6, x7 `plus` y7)
+  times (x1, x2, x3, x4, x5, x6, x7) (y1, y2, y3, y4, y5, y6, y7) =
+    (x1 `times` y1, x2 `times` y2, x3 `times` y3, x4 `times` y4, x5 `times` y5, x6 `times` y6, x7 `times` y7)
+
+instance (Ring a, Ring b) => Ring (a,b) where
+  negate (x1, x2) = (negate x1, negate x2)
+
+instance (Ring a, Ring b, Ring c) => Ring (a,b,c) where
+  negate (x1, x2, x3) = (negate x1, negate x2, negate x3)
+
+instance (Ring a, Ring b, Ring c, Ring d) => Ring (a,b,c,d) where
+  negate (x1, x2, x3, x4) = (negate x1, negate x2, negate x3, negate x4)
+
+instance (Ring a, Ring b, Ring c, Ring d, Ring e) => Ring (a,b,c,d,e) where
+  negate (x1, x2, x3, x4, x5) = (negate x1, negate x2, negate x3, negate x4, negate x5)
+
+instance (Ring a, Ring b, Ring c, Ring d, Ring e, Ring f) => Ring (a,b,c,d,e,f) where
+  negate (x1, x2, x3, x4, x5, x6) = (negate x1, negate x2, negate x3, negate x4, negate x5, negate x6)
+
+instance (Ring a, Ring b, Ring c, Ring d, Ring e, Ring f, Ring g) => Ring (a,b,c,d,e,f,g) where
+  negate (x1, x2, x3, x4, x5, x6, x7) = (negate x1, negate x2, negate x3, negate x4, negate x5, negate x6, negate x7)
+
 instance Semiring b => Semiring (a -> b) where
   plus f g    = \x -> f x `plus` g x
   zero        = const zero
@@ -926,22 +1027,48 @@ deriving via (WrappedNum CSChar) instance Semiring CSChar
 deriving via (WrappedNum CChar) instance Semiring CChar
 deriving via (WrappedNum IntPtr) instance Semiring IntPtr
 deriving via (WrappedNum WordPtr) instance Semiring WordPtr
-#if !HOST_OS_WINDOWS
+
+#ifdef HTYPE_CC_T
 deriving via (WrappedNum CCc) instance Semiring CCc
-deriving via (WrappedNum CDev) instance Semiring CDev
-deriving via (WrappedNum CGid) instance Semiring CGid
-deriving via (WrappedNum CIno) instance Semiring CIno
-deriving via (WrappedNum CMode) instance Semiring CMode
-deriving via (WrappedNum CNlink) instance Semiring CNlink
-deriving via (WrappedNum COff) instance Semiring COff
-deriving via (WrappedNum CPid) instance Semiring CPid
-deriving via (WrappedNum CRLim) instance Semiring CRLim
-deriving via (WrappedNum CSpeed) instance Semiring CSpeed
-deriving via (WrappedNum CSsize) instance Semiring CSsize
-deriving via (WrappedNum CTcflag) instance Semiring CTcflag
-deriving via (WrappedNum CUid) instance Semiring CUid
-deriving via (WrappedNum Fd) instance Semiring Fd
 #endif
+#ifdef HTYPE_DEV_T
+deriving via (WrappedNum CDev) instance Semiring CDev
+#endif
+#ifdef HTYPE_GID_T
+deriving via (WrappedNum CGid) instance Semiring CGid
+#endif
+#ifdef HTYPE_INO_T
+deriving via (WrappedNum CIno) instance Semiring CIno
+#endif
+#ifdef HTYPE_MODE_T
+deriving via (WrappedNum CMode) instance Semiring CMode
+#endif
+#ifdef HTYPE_NLINK_T
+deriving via (WrappedNum CNlink) instance Semiring CNlink
+#endif
+#ifdef HTYPE_OFF_T
+deriving via (WrappedNum COff) instance Semiring COff
+#endif
+#ifdef HTYPE_PID_T
+deriving via (WrappedNum CPid) instance Semiring CPid
+#endif
+#ifdef HTYPE_RLIM_T
+deriving via (WrappedNum CRLim) instance Semiring CRLim
+#endif
+#ifdef HTYPE_SPEED_T
+deriving via (WrappedNum CSpeed) instance Semiring CSpeed
+#endif
+#ifdef HTYPE_SSIZE_T
+deriving via (WrappedNum CSsize) instance Semiring CSsize
+#endif
+#ifdef HTYPE_TCFLAG_T
+deriving via (WrappedNum CTcflag) instance Semiring CTcflag
+#endif
+#ifdef HTYPE_UID_T
+deriving via (WrappedNum CUid) instance Semiring CUid
+#endif
+deriving via (WrappedNum Fd) instance Semiring Fd
+
 deriving via (WrappedNum Natural) instance Semiring Natural
 #else
 -- Integral and fieldlike instances
@@ -1002,22 +1129,48 @@ $(let
    ,[t|CChar|]
    ,[t|IntPtr|]
    ,[t|WordPtr|]
-#if !HOST_OS_WINDOWS
+
+#ifdef HTYPE_CC_T
    ,[t|CCc|]
-   ,[t|CDev|]
-   ,[t|CGid|]
-   ,[t|CIno|]
-   ,[t|CMode|]
-   ,[t|CNlink|]
-   ,[t|COff|]
-   ,[t|CPid|]
-   ,[t|CRLim|]
-   ,[t|CSpeed|]
-   ,[t|CSsize|]
-   ,[t|CTcflag|]
-   ,[t|CUid|]
-   ,[t|Fd|]
 #endif
+#ifdef HTYPE_DEV_T
+   ,[t|CDev|]
+#endif
+#ifdef HTYPE_GID_T
+   ,[t|CGid|]
+#endif
+#ifdef HTYPE_INO_T
+   ,[t|CIno|]
+#endif
+#ifdef HTYPE_MODE_T
+   ,[t|CMode|]
+#endif
+#ifdef HTYPE_NLINK_T
+   ,[t|CNlink|]
+#endif
+#ifdef HTYPE_OFF_T
+   ,[t|COff|]
+#endif
+#ifdef HTYPE_PID_T
+   ,[t|CPid|]
+#endif
+#ifdef HTYPE_RLIM_T
+   ,[t|CRLim|]
+#endif
+#ifdef HTYPE_SPEED_T
+   ,[t|CSpeed|]
+#endif
+#ifdef HTYPE_SSIZE_T
+   ,[t|CSsize|]
+#endif
+#ifdef HTYPE_TCFLAG_T
+   ,[t|CTcflag|]
+#endif
+#ifdef HTYPE_UID_T
+   ,[t|CUid|]
+#endif
+   ,[t|Fd|]
+
    ,[t|Natural|]
    ])
 #endif
@@ -1064,22 +1217,46 @@ deriving via (WrappedNum CChar) instance Ring CChar
 deriving via (WrappedNum IntPtr) instance Ring IntPtr
 deriving via (WrappedNum WordPtr) instance Ring WordPtr
 
-#if !HOST_OS_WINDOWS
+#ifdef HTYPE_CC_T
 deriving via (WrappedNum CCc) instance Ring CCc
-deriving via (WrappedNum CDev) instance Ring CDev
-deriving via (WrappedNum CGid) instance Ring CGid
-deriving via (WrappedNum CIno) instance Ring CIno
-deriving via (WrappedNum CMode) instance Ring CMode
-deriving via (WrappedNum CNlink) instance Ring CNlink
-deriving via (WrappedNum COff) instance Ring COff
-deriving via (WrappedNum CPid) instance Ring CPid
-deriving via (WrappedNum CRLim) instance Ring CRLim
-deriving via (WrappedNum CSpeed) instance Ring CSpeed
-deriving via (WrappedNum CSsize) instance Ring CSsize
-deriving via (WrappedNum CTcflag) instance Ring CTcflag
-deriving via (WrappedNum CUid) instance Ring CUid
-deriving via (WrappedNum Fd) instance Ring Fd
 #endif
+#ifdef HTYPE_DEV_T
+deriving via (WrappedNum CDev) instance Ring CDev
+#endif
+#ifdef HTYPE_GID_T
+deriving via (WrappedNum CGid) instance Ring CGid
+#endif
+#ifdef HTYPE_INO_T
+deriving via (WrappedNum CIno) instance Ring CIno
+#endif
+#ifdef HTYPE_MODE_T
+deriving via (WrappedNum CMode) instance Ring CMode
+#endif
+#ifdef HTYPE_NLINK_T
+deriving via (WrappedNum CNlink) instance Ring CNlink
+#endif
+#ifdef HTYPE_OFF_T
+deriving via (WrappedNum COff) instance Ring COff
+#endif
+#ifdef HTYPE_PID_T
+deriving via (WrappedNum CPid) instance Ring CPid
+#endif
+#ifdef HTYPE_RLIM_T
+deriving via (WrappedNum CRLim) instance Ring CRLim
+#endif
+#ifdef HTYPE_SPEED_T
+deriving via (WrappedNum CSpeed) instance Ring CSpeed
+#endif
+#ifdef HTYPE_SSIZE_T
+deriving via (WrappedNum CSsize) instance Ring CSsize
+#endif
+#ifdef HTYPE_TCFLAG_T
+deriving via (WrappedNum CTcflag) instance Ring CTcflag
+#endif
+#ifdef HTYPE_UID_T
+deriving via (WrappedNum CUid) instance Ring CUid
+#endif
+deriving via (WrappedNum Fd) instance Ring Fd
 #else
 $(let
   deriveRing :: Q Type -> Q [Dec]
@@ -1131,21 +1308,45 @@ $(let
     ,[t|IntPtr|]
     ,[t|WordPtr|]
 
-#if !HOST_OS_WINDOWS
+#ifdef HTYPE_CC_T
     ,[t|CCc|]
-    ,[t|CDev|]
-    ,[t|CGid|]
-    ,[t|CIno|]
-    ,[t|CMode|]
-    ,[t|CNlink|]
-    ,[t|COff|]
-    ,[t|CPid|]
-    ,[t|CRLim|]
-    ,[t|CSpeed|]
-    ,[t|CSsize|]
-    ,[t|CTcflag|]
-    ,[t|CUid|]
-    ,[t|Fd|]
 #endif
+#ifdef HTYPE_DEV_T
+    ,[t|CDev|]
+#endif
+#ifdef HTYPE_GID_T
+    ,[t|CGid|]
+#endif
+#ifdef HTYPE_INO_T
+    ,[t|CIno|]
+#endif
+#ifdef HTYPE_MODE_T
+    ,[t|CMode|]
+#endif
+#ifdef HTYPE_NLINK_T
+    ,[t|CNlink|]
+#endif
+#ifdef HTYPE_OFF_T
+    ,[t|COff|]
+#endif
+#ifdef HTYPE_PID_T
+    ,[t|CPid|]
+#endif
+#ifdef HTYPE_RLIM_T
+    ,[t|CRLim|]
+#endif
+#ifdef HTYPE_SPEED_T
+    ,[t|CSpeed|]
+#endif
+#ifdef HTYPE_SSIZE_T
+    ,[t|CSsize|]
+#endif
+#ifdef HTYPE_TCFLAG_T
+    ,[t|CTcflag|]
+#endif
+#ifdef HTYPE_UID_T
+    ,[t|CUid|]
+#endif
+    ,[t|Fd|]
     ])
 #endif
